import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.util.*;
import java.awt.Color;
import java.awt.Point;

/**
 * A ListBox
 * 
 * @author Taylor Born
 * @version February 2011
 */
public class ListBox  extends Actor
{
    // Size of this ListBox
    private Point size;
    
    // Height of scrollBar
    private int scrollBar = 40;
    
    // Y position of where mouse pressed on scrollBar, used as reference for when dragged
    private int mouseY;
    
    private int textHeight = 14;
    private ArrayList<String> strs = new ArrayList<String>();
    private int scroll;
    private int index;
    private boolean dragging;
    private Point lastMouse = new Point(-25, -25);
    
    /**
     * Create a new ListBox.
     * @param size The width and height of the ListBox.
     * @param strings The list of Strings to display in the ListBox.
     */
    public ListBox(Point size, ArrayList<String> strings)
    {
        this.size = size;
        strs = strings;
    }
    
    private void paint()
    {
        GreenfootImage pic = new GreenfootImage((int)size.getX(), (int)size.getY());
        pic.setColor(Color.WHITE);
        pic.fill();
        int scrollValue = (int)(((double)scroll / (size.getY() - scrollBar)) * (strs.size() * textHeight - size.getY() + 1));
        for (int i = 0; i < strs.size(); i++)
        {
            if (index == i)
            {
                pic.setColor(Color.YELLOW);
                pic.fillRect(0, textHeight * i - scrollValue, (int)size.getX() - 10, textHeight);
            }
            else if (!dragging && lastMouse.getX() - (getX() - size.getX() / 2) > 0 && lastMouse.getX() - (getX() - size.getX() / 2) < size.getX() - 10 && lastMouse.getY() - (getY() - size.getY() / 2) > 0 && lastMouse.getY() - (getY() - size.getY() / 2) < size.getY() && lastMouse.getY() - (getY() - size.getY() / 2) > textHeight * i - scrollValue && lastMouse.getY() - (getY() - size.getY() / 2) <= textHeight * i - scrollValue + textHeight)
            {
                pic.setColor(new Color(185, 211, 225));
                pic.fillRect(0, textHeight * i - scrollValue, (int)size.getX() - 10, textHeight);
            }
            pic.setColor(Color.BLACK);
            pic.drawRect(0, textHeight * i - scrollValue, (int)size.getX() - 10, textHeight);
            pic.drawString(strs.get(i), 4, textHeight * i - scrollValue + textHeight - 2);
        }
        if (strs.size() > size.getY() / textHeight)
        {
            pic.setColor(Color.RED);
            pic.fillRect((int)size.getX() - 9, scroll, 9, scrollBar);
        }
        pic.setColor(Color.BLACK);
        pic.drawRect(0, 0, (int)size.getX() - 1, (int)size.getY() - 1);
        setImage(pic);
    }
    
    public void act()
    {
        MouseInfo mouse = Greenfoot.getMouseInfo();
        if (strs.size() > size.getY() / textHeight && Greenfoot.mousePressed(this) && mouse.getX() - (getX() - size.getX() / 2) > size.getX() - 10 && mouse.getX() - (getX() - size.getX() / 2) < size.getX() && mouse.getY() - (getY() - size.getY() / 2) > scroll && mouse.getY() - (getY() - size.getY() / 2) < scrollBar + scroll)
        {
            dragging = true;
            mouseY = (int)mouse.getY() - (getY() - (int)size.getY() / 2) - scroll;
        }
        if (Greenfoot.mouseClicked(null) || Greenfoot.mouseDragEnded(null))
            dragging = false;
        if (dragging && Greenfoot.mouseDragged(null))
        {
            scroll += mouse.getY() - (getY() - size.getY() / 2) - scroll - mouseY;
            if (scroll < 0)
                scroll = 0;
            else if (scroll + scrollBar > size.getY())
                scroll = (int)size.getY() - scrollBar;
        }
        if (Greenfoot.mouseDragged(null) || Greenfoot.mouseMoved(null))
            lastMouse = new Point(mouse.getX(), mouse.getY());
        if (Greenfoot.mousePressed(this) && mouse.getX() - (getX() - size.getX() / 2) > 0 && mouse.getX() - (getX() - size.getX() / 2) < size.getX() - 10 && mouse.getY() - (getY() - size.getY() / 2) > 0 && mouse.getY() - (getY() - size.getY() / 2) < size.getY())
        {
            for (int i = 0; i < strs.size(); i++)
                if (mouse.getY() - (getY() - size.getY() / 2) > textHeight * i - (int)(((double)scroll / (size.getY() - scrollBar)) * (strs.size() * textHeight - size.getY() + 1)) && mouse.getY() - (getY() - size.getY() / 2) <= textHeight * i - (int)(((double)scroll / (size.getY() - scrollBar)) * (strs.size() * textHeight - size.getY() + 1)) + textHeight)
                    index = i;
        }
        scrollBar = (int)(((double)size.getY() / (strs.size() * textHeight + 1)) * size.getY());
        if (scrollBar < 5)
            scrollBar = 5;
        paint();
    }
    
    public void addedToWorld(World world)
    {
        scroll = 0;
        index = -1;
        paint();
    }
    
    /**
     * Get the index of what is selected.
     * @return The index of what is selected. -1 if none selected.
     */
    public int getIndex()
    {
        return index;
    }
    
    /**
     * Set which item to be selected.
     * @param index The index of the item to be selected.
     */
    public void setIndex(int index)
    {
        // Check to see if valid index
        if (index > -1 && index < strs.size())
            this.index = index;
        else
            this.index = -1;
    }
    
    /**
     * Get the selected String.
     * @return The selected String.
     */
    public String getSelected()
    {
        // Check to make sure a selection has been made
        if (index == -1)
            return null;
        return strs.get(index);
    }
    
    /**
     * Get the String in the ListBox at the given index.
     * @param index The index for what String in the ListBox to get.
     * @return The String at the index of the ListBox.
     */
    public String getString(int index)
    {
        // Check to see if valid index
        if (index > -1 && index < strs.size())
            return strs.get(index);
        return null;
    }
    
    /**
     * Get the contents of the ListBox.
     * @return The contents of the ListBox.
     */
    public ArrayList<String> getList()
    {
        return strs;
    }
    
    /**
     * Add a String to the ListBox, making it selected and in view.
     * @param s The String to be added.
     */
    public void addString(String s)
    {
        strs.add(s);
        index = strs.size() - 1;
        scrollBar = (int)(((double)size.getY() / (strs.size() * textHeight + 1)) * size.getY());
        if (scrollBar < 5)
            scrollBar = 5;
        if (strs.size() * textHeight - 2 > size.getY())
            scroll = (int)size.getY() - scrollBar;
    }
    
    /**
     * Remove what is selected.
     */
    public void removeSelected()
    {
        if (index != -1)
        {
            strs.remove(index);
            index = -1;
        }
    }
    
    /**
     * Get the number of items in the ListBox.
     * @return Number of items in the ListBox.
     */
    public int listSize()
    {
        return strs.size();
    }
    
    /**
     * Remove everything from the ListBox.
     */
    public void clear()
    {
        strs.clear();
        index = -1;
    }
    
    /**
     * Change the String at a given index.
     * @param index The index in the ListBox to change.
     * @param string The String to set to at the given index.
     */
    public void set(int index, String string)
    {
        if (index < strs.size())
            strs.set(index, string);
    }
    
    public void scrollToTop()
    {
        scroll = 0;
    }
}