import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.util.*;
import java.awt.Point;

/**
 * A TextBox
 * 
 * @author Taylor Born
 * @version November 2010
 */
public class TextBox  extends Actor
{
    private boolean hasFocus;
    private String text = "";
    private Point size;
    private Font font = new Font("New Times Roman", Font.PLAIN, 12);
    private Point clickedAt;
    private int caret;
    private int blink;
    private ArrayList<String> strs;
    private int scroll;
    private boolean dragging;
    private Point lastMouse = new Point(-25, -25);
    private int[] presses = new int[4];
    private String accept = "";
    private String dontAccept = "";
    private int maxLength = -1;
    
    private boolean readOnly;
    
    private int scrollBar = 40;
    private int mouseY;
    
    private Color textColor = Color.BLACK;
    private Color blinkColor = Color.BLUE;
    private Color activeColor = Color.RED;
    private Color scrollColor = Color.BLUE;
    private Color backColor = Color.WHITE;
    
    private final int SCROLL_WIDTH = 9;
    
    private boolean hasChanged = false;
    
    private Integer lastCursorX = null;
    
    /**
     * Create a new TextBox.
     * @param size The width and height of the TextBox.
     * @param text The initial text to display in the TextBox.
     */
    public TextBox(Point size, String text)
    {
        this.size = size;
        this.text = text;
        caret = text.length();
        paint(false);
    }
    
    public void act() 
    {
        if (blink++ == 20)
            blink = -20;
        MouseInfo mouse = Greenfoot.getMouseInfo();
        if (!Greenfoot.mouseDragEnded(null))
            if (Greenfoot.mousePressed(this) || Greenfoot.mouseClicked(this))
            {
                if (!hasFocus)
                {
                    Greenfoot.getKey();
                    hasFocus = true;
                }
                blink = 1;
                if (mouse.getX() - (getX() - size.getX() / 2) < size.getX() - SCROLL_WIDTH - 3) // -20
                {
                    clickedAt = new Point(mouse.getX() - (getX() - (int)size.getX() / 2), mouse.getY() - (getY() - (int)size.getY() / 2));
                    lastCursorX = null;
                }
                else if (Greenfoot.mouseClicked(this))
                    dragging = false;
                else if (strs.size() * font.getSize() > (int)size.getY() && mouse.getX() - (getX() - size.getX() / 2) > (int)size.getX() - 15 && mouse.getY() - (getY() - (int)size.getY() / 2) > scroll && mouse.getY() - (getY() - (int)size.getY() / 2) < scroll + scrollBar)
                {
                    dragging = true;
                    mouseY = mouse.getY() - (getY() - (int)size.getY() / 2) - scroll;
                }
            }
            else if (Greenfoot.mousePressed(null) || Greenfoot.mouseClicked(null))
                hasFocus = false;
        if (Greenfoot.mouseDragEnded(null))
            dragging = false;
        int oldCaret = caret;
        if (hasFocus)
        {
            String s = Greenfoot.getKey();
            Graphics2D graphics = (new GreenfootImage(1, 1)).getAwtImage().createGraphics();
            graphics.setFont(font);
            FontMetrics fm = graphics.getFontMetrics();
            int scrollValue = strs.size() * font.getSize() > size.getY() ? (int)(((double)scroll / ((double)size.getY() - scrollBar)) * (strs.size() * font.getSize() - size.getY() + 3)) : 0;
            if (s != null && !readOnly)
            {
                if (s.equals("backspace"))
                {
                    if (caret > 0)
                    {
                        text = text.substring(0, caret - 1) + text.substring(caret);
                        caret--;
                        blink = 1;
                    }
                }
                else if (s.equals("left") || s.equals("right") || s.equals("up") || s.equals("down"))
                {}
                else
                {
                    if (s.equals("space"))
                        s = " ";
                    else if (s.equals("enter"))
                        s = "\n";
                    if (s.length() == 1 && (maxLength == -1 || text.length() < maxLength) && (accept.equals("") || accept.contains(s)) && !dontAccept.contains(s))
                    {
                        text = text.substring(0, caret) + s + text.substring(caret);
                        caret++;
                    }
                }
                graphics.dispose();
            }
            if (Greenfoot.isKeyDown("up"))
            {
                presses[0]++;
                if (presses[0] == 25)
                {
                    if (caret >= strs.get(0).length())
                    {
                        int l = 0;
                        for (int i = 0; i < strs.size(); i++)
                        {
                            if (l + strs.get(i).length() > caret)
                            {
                                int cur;
                                if (lastCursorX == null)
                                {
                                    cur = 4 + fm.charsWidth(strs.get(i).substring(0, caret - l).toCharArray(), 0, caret - l);
                                    lastCursorX = cur;
                                }
                                else
                                    cur = lastCursorX;
                                clickedAt = new Point(cur, i * font.getSize() - scrollValue);
                                blink = 1;
                                break;
                            }
                            l += strs.get(i).length();
                        }
                    }
                    else
                        caret = 0;
                    presses[0] = 20;
                }
            }
            else
            {
                if (presses[0] > 0 && presses[0] < 20)
                {
                    if (caret >= strs.get(0).length())
                    {
                        int l = 0;
                        for (int i = 0; i < strs.size(); i++)
                        {
                            if (l + strs.get(i).length() > caret)
                            {
                                int cur;
                                if (lastCursorX == null)
                                {
                                    cur = 4 + fm.charsWidth(strs.get(i).substring(0, caret - l).toCharArray(), 0, caret - l);
                                    lastCursorX = cur;
                                }
                                else
                                    cur = lastCursorX;
                                //int cur = 4 + fm.charsWidth(strs.get(i).substring(0, caret - l).toCharArray(), 0, caret - l);
                                clickedAt = new Point(cur, i * font.getSize() - scrollValue);
                                blink = 1;
                                break;
                            }
                            l += strs.get(i).length();
                        }
                    }
                    else
                        caret = 0;
                }
                presses[0] = 0;
            }
            if (Greenfoot.isKeyDown("down"))
            {
                presses[1]++;
                if (presses[1] == 25)
                {
                    int l = 0;
                    for (int i = 0; i < strs.size(); i++)
                    {
                        if (l + strs.get(i).length() > caret)
                        {
                            int cur;
                            if (lastCursorX == null)
                            {
                                cur = 4 + fm.charsWidth(strs.get(i).substring(0, caret - l).toCharArray(), 0, caret - l);
                                lastCursorX = cur;
                            }
                            else
                                cur = lastCursorX;
                            clickedAt = new Point(cur, (i + 2) * font.getSize() - scrollValue - 2);
                            blink = 1;
                            break;
                        }
                        l += strs.get(i).length();
                    }
                    presses[1] = 20;
                }
            }
            else
            {
                if (presses[1] > 0 && presses[1] < 20)
                {
                    int l = 0;
                    for (int i = 0; i < strs.size(); i++)
                    {
                        if (l + strs.get(i).length() > caret)
                        {
                            int cur;
                            if (lastCursorX == null)
                            {
                                cur = 4 + fm.charsWidth(strs.get(i).substring(0, caret - l).toCharArray(), 0, caret - l);
                                lastCursorX = cur;
                            }
                            else
                                cur = lastCursorX;
                            clickedAt = new Point(cur, (i + 2) * font.getSize() - scrollValue - 2);
                            blink = 1;
                            break;
                        }
                        l += strs.get(i).length();
                    }
                }
                presses[1] = 0;
            }
            if (Greenfoot.isKeyDown("left"))
            {
                presses[2]++;
                if (presses[2] == 25)
                {
                    if (caret > 0)
                        caret--;
                    blink = 1;
                    presses[2] = 20;
                }
            }
            else
            {
                if (presses[2] > 0 && presses[2] < 20)
                {
                    if (caret > 0)
                        caret--;
                    blink = 1;
                }
                presses[2] = 0;
            }
            if (Greenfoot.isKeyDown("right"))
            {
                presses[3]++;
                if (presses[3] == 25)
                {
                    if (caret < text.length())
                        caret++;
                    blink = 1;
                    presses[3] = 20;
                }
            }
            else
            {
                if (presses[3] > 0 && presses[3] < 20)
                {
                    if (caret < text.length())
                        caret++;
                    blink = 1;
                }
                presses[3] = 0;
            }
            if (Greenfoot.mouseDragged(null) || Greenfoot.mouseMoved(null))
            {
                if (dragging)
                    scroll += mouse.getY() - (getY() - size.getY() / 2) - scroll - mouseY;
            }
        }
        if (Greenfoot.mouseDragged(null) || Greenfoot.mouseMoved(null))
            lastMouse = new Point(mouse.getX(), mouse.getY());
        scrollBar = (int)(((double)size.getY() / (strs.size() * font.getSize() + 1)) * size.getY());
        if (scrollBar < 5)
            scrollBar = 5;
        if (scroll < 0)
            scroll = 0;
        else if (scroll + scrollBar > size.getY())
            scroll = (int)size.getY() - scrollBar;
        paint(oldCaret != caret);
    }
    
    private void paint(boolean changed)
    {
        if (changed)
        {
            lastCursorX = null;
            hasChanged = true;
        }
        GreenfootImage pic = new GreenfootImage((int)size.getX(), (int)size.getY());
        pic.setColor(backColor);
        pic.fill();
        
        pic.setFont(font);
        Graphics2D graphics = pic.getAwtImage().createGraphics();
        graphics.setFont(font);
        FontMetrics fm = graphics.getFontMetrics();
        String s = text + " ";
        String t = "";
        int line = 0;
        int count = 0;
        int strSize = strs == null ? 0 : strs.size();
        int scrollValue = strSize * font.getSize() > size.getY() ? (int)(((double)scroll / ((double)size.getY() - scrollBar)) * (strSize * font.getSize() - size.getY() + 3)) : 0;
        
        int caretLine = -1;
        
        strs = new ArrayList<String>();
        
        if (clickedAt != null)
        {
            String strr = "";
        }
        
        String ss = "";
        if (s.indexOf("\n") != -1)
        {
            ss = s.substring(s.indexOf("\n") + 1);
            s = s.substring(0, s.indexOf("\n")) + " ";
        }
        while (s.length() > 0 || ss.length() > 0)
        {
            while (!s.equals(""))
            {
                int width = fm.charsWidth(t.toCharArray(), 0, t.length());
                int newW = fm.charsWidth(s.substring(0, s.indexOf(" ") + 1).toCharArray(), 0, s.substring(0, s.indexOf(" ") + 1).length());
                
                if (s.length() == 0 || width + newW > size.getX() - SCROLL_WIDTH - 3)
                {
                    if (t.length() == 0)
                        for (int i = 1; i <= s.substring(0, s.indexOf(" ") + 1).length(); i++)
                        {
                            if (fm.charsWidth(s.substring(0, i).toCharArray(), 0, s.substring(0, i).length()) > size.getX() - SCROLL_WIDTH - 3) // -24
                            {
                                t = s.substring(0, i - 1);
                                s = s.substring(i - 1);
                                break;
                            }
                        }
                    pic.setColor(textColor);
                    pic.drawString(t, 4, font.getSize() * ++line - scrollValue);
                    strs.add(t);
                    if (clickedAt != null && clickedAt.getY() > font.getSize() * (line - 1) - scrollValue && clickedAt.getY() <= font.getSize() * line - scrollValue)
                    {
                        if (clickedAt.getX() < 4 + fm.charsWidth(t.substring(0, 1).toCharArray(), 0, 1) / 2)
                        {
                            caret = count;
                            clickedAt = null;
                        }
                        else
                        {
                            for (int i = 1; i < t.length() - 1; i++)
                                if (clickedAt.getX() >= 4 + fm.charsWidth(t.substring(0, i - 1).toCharArray(), 0, i - 1) + fm.charsWidth(t.substring(i - 1, i).toCharArray(), 0, 1) / 2 && clickedAt.getX() < 4 + fm.charsWidth(t.substring(0, i).toCharArray(), 0, i) + fm.charsWidth(t.substring(i, i + 1).toCharArray(), 0, 1) / 2)
                                {
                                    caret = count + i;
                                    clickedAt = null;
                                    break;
                                }
                            if (clickedAt != null)
                            {
                                caret = count + t.length() - 1;
                                clickedAt = null;
                            }
                        }
                        changed = true;
                        caretLine = line;
                    }
                    if (hasFocus && caret >= count && caret < count + t.length())
                    {
                        if (blink > 0)
                        {
                            pic.setColor(blinkColor);
                            int n = fm.charsWidth(t.substring(0, caret - count).toCharArray(), 0, caret - count) + 4;
                            pic.fillRect(n, font.getSize() * (line - 1) + font.getSize() / 5 - scrollValue, 2, font.getSize());
                        }
                        caretLine = line;
                    }
                    count += t.length();
                    t = "";
                }
                if (fm.charsWidth(s.substring(0, s.indexOf(" ") + 1).toCharArray(), 0, s.substring(0, s.indexOf(" ") + 1).length()) > size.getX() - SCROLL_WIDTH - 3) // -24
                {
                    for (int i = 1; i <= s.substring(0, s.indexOf(" ") + 1).length(); i++)
                    {
                        if (fm.charsWidth(s.substring(0, i).toCharArray(), 0, s.substring(0, i).length()) > size.getX() - SCROLL_WIDTH - 3) // -24
                        {
                            t = s.substring(0, i - 1);
                            s = s.substring(i - 1);
                            break;
                        }
                    }
                }
                else
                {
                    t += s.substring(0, s.indexOf(" ") + 1);
                    s = s.substring(s.indexOf(" ") + 1);
                }
            }
            pic.setColor(textColor);
            pic.drawString(t, 4, font.getSize() * ++line - scrollValue);
            strs.add(t);
            
            if (clickedAt != null && clickedAt.getY() >= font.getSize() * (line - 1) - scrollValue && clickedAt.getY() <= font.getSize() * line - scrollValue)
            {
                if (clickedAt.getX() < 4 + fm.charsWidth(t.substring(0, 1).toCharArray(), 0, 1) / 2)
                {
                    caret = count;
                    clickedAt = null;
                }
                else
                {
                    for (int i = 1; i < t.length() - 1; i++)
                        if (clickedAt.getX() >= 4 + fm.charsWidth(t.substring(0, i - 1).toCharArray(), 0, i - 1) + fm.charsWidth(t.substring(i - 1, i).toCharArray(), 0, 1) / 2 && clickedAt.getX() < 4 + fm.charsWidth(t.substring(0, i).toCharArray(), 0, i) + fm.charsWidth(t.substring(i, i + 1).toCharArray(), 0, 1) / 2)
                        {
                            caret = count + i;
                            clickedAt = null;
                            break;
                        }
                    if (clickedAt != null)
                    {
                        caret = count + t.length() - (ss.length() == 0 ? 2 : 1);
                        clickedAt = null;
                    }
                }
                changed = true;
                caretLine = line;
            }
            if (hasFocus && caret >= count && caret < count + t.length())
            {
                if (blink > 0)
                {
                    pic.setColor(blinkColor);
                    int n = fm.charsWidth(t.substring(0, caret - count).toCharArray(), 0, caret - count) + 4;
                    pic.fillRect(n, font.getSize() * (line - 1) + font.getSize() / 5 - scrollValue, 2, font.getSize());
                }
                caretLine = line;
            }
            count += t.length();
            t = "";
            if (ss.length() > 0)
            {
                if (ss.indexOf("\n") != -1)
                {
                    s = ss.substring(0, ss.indexOf("\n")) + " ";
                    ss = ss.substring(ss.indexOf("\n") + 1);
                }
                else
                {
                    s = ss + " ";
                    ss = "";
                }
            }
        }
        if (clickedAt != null)
        {
            caret = text.length();
            clickedAt = null;
        }
        
        graphics.dispose();
        
        if (changed && caretLine != -1)
        {
            int y = font.getSize() * (caretLine - 1) + font.getSize() / 5 - scrollValue;
            boolean c = true;
            if (y < 0)
                scrollValue = font.getSize() * (caretLine - 1) + font.getSize() / 5;
            else if (y + font.getSize() > size.getY())
                scrollValue = font.getSize() * (caretLine - 1) + font.getSize() / 5 + font.getSize() - (int)size.getY() + 3;
            else
                c = false;
            if (c)
                scroll = (int)((double)scrollValue / (strSize * font.getSize() - size.getY() + 3) * ((double)size.getY() - scrollBar));
        }
        
        if (strs.size() * font.getSize() > size.getY())
        {
            pic.setColor(scrollColor);
            pic.fillRect((int)size.getX() - SCROLL_WIDTH, scroll, SCROLL_WIDTH, scrollBar);
            pic.setColor(Color.BLACK);
            pic.drawRect((int)size.getX() - SCROLL_WIDTH, scroll, SCROLL_WIDTH, scrollBar);
        }
        
        pic.setColor(Color.BLACK);
        pic.drawRect(0, 0, (int)size.getX() - 1, (int)size.getY() - 1);
        if (hasFocus)
        {
            pic.setColor(activeColor);
            pic.drawRect(1, 1, (int)size.getX() - 2, (int)size.getY() - 2);
        }
        
        setImage(pic);
    }
    
    /**
     * Get the contents of this TextBox.
     * @return The text in this TextBox.
     */
    public String getText()
    {
        return text;
    }
    
    /**
     * Set the text for this TextBox.
     * @param str The text to be in this TextBox.
     */
    public void setText(String str)
    {
        text = str;
        if (caret > str.length())
        {
            caret = str.length();
            lastCursorX = null;
        }
        paint(true);
    }
    
    /**
     * Get the Font being used in this TextBox.
     * @return The Font being used in this TextBox.
     */
    public Font getFont()
    {
        return font;
    }
    
    /**
     * Set the Font for this TextBox.
     * @param f The Font to set for this TextBox.
     */
    public void setFont(Font f)
    {
        font = (f);
    }
    
    /**
     * Set String of characters that will be allowed as input.
     * @param s String of characters that will be allowed as input. Having an empty String will allow all characters to be inputed.
     */
    public void acceptOnly(String s)
    {
        accept = s;
    }
    
    /**
     * Set String of characters that will not be allowed as input.
     * @param s String of characters that will not be allowed as input.
     */
    public void dontAccept(String s)
    {
        dontAccept = s;
    }
    
    /**
     * Set a limit for how long the text can be.
     * @param m The limit for how many characters the text can include. -1 will allow unlimited.
     */
    public void setMaxLength(int m)
    {
        maxLength = m;
    }
    
    
    public void addedToWorld(World world)
    {
        hasFocus = false;
    }
    
    /**
     * Determine whether this TextBox has focus.
     * @return Whether this TextBox has focus or not.
     */
    public boolean focused()
    {
        return hasFocus;
    }
    
    /**
     * Set the width and height of this TextBox.
     * @param s The width and height for this TextBox.
     */
    public void setSize(Point s)
    {
        size = s;
    }
    
    /**
     * Set whether the TextBox will accept input.
     * @param readOnly Whether the TextBox will accept input.
     */
    public void setReadOnly(boolean readOnly)
    {
        this.readOnly = readOnly;
    }
    
    /**
     * Determine whether this TextBox will accept input.
     * @return Whether this TextBox accepts input.
     */
    public boolean isReadOnly()
    {
        return readOnly;
    }
    
    /**
     * Get the Color this TextBox will be painting its text as.
     * @return The Color this TextBox will be painting its text as.
     */
    public Color getTextColor()
    {
        return textColor;
    }
    
    /**
     * Change the Color this TextBox will be painting its text as.
     * @param c The Color this TextBox will be painting its text as.
     */
    public void setTextColor(Color c)
    {
        textColor = c;
    }
    
    /**
     * Get the Color of the blinking cursor.
     * @return The Color of the blinking cursor.
     */
    public Color getCursorColor()
    {
        return blinkColor;
    }
    
    /**
     * Change the Color of the blinking cursor.
     * @param c The new Color of the blinking cursor.
     */
    public void setCursorColor(Color c)
    {
        blinkColor = c;
    }
    
    /**
     * Get the Color outlining the TextBox when it has focus.
     * @return The Color outlining the TextBox when it has focus.
     */
    public Color getOutlineColor()
    {
        return activeColor;
    }
    
    /**
     * Change the Color outlining the TextBox when it has focus.
     * @param c The new Color outlining the TextBox when it has focus.
     */
    public void setOutlineColor(Color c)
    {
        activeColor = c;
    }
    
    /**
     * Get the Color of this TextBox's scroll bar.
     * @return The Color of this TextBox's scroll bar.
     */
    public Color getScrollBarColor()
    {
        return scrollColor;
    }
    
    /**
     * Change the Color of this TextBox's scroll bar.
     * @param c The new Color of this TextBox's scroll bar.
     */
    public void setScrollBarColor(Color c)
    {
        scrollColor = c;
    }
    
    /**
     * Get the Color of this TextBox's background.
     * @return The Color of this TextBox's background.
     */
    public Color getBackgroundColor()
    {
        return backColor;
    }
    
    /**
     * Change the Color of this TextBox's background.
     * @param c The new Color of this TextBox's background.
     */
    public void setBackgroundColor(Color c)
    {
        backColor = c;
    }
    
    public boolean hasChanged()
    {
        boolean c = hasChanged;
        hasChanged = false;
        return c;
    }
}